/**
 * A minimal plain text editor with syntax highlighting, line numbers, and line highlighting.
 *    `<light-editor>` is not intended to replace full solutions like CodeMirror, but it
 *    is a lightweight alternative using a `<pre>` overlaid on top of a `<textarea>`
 *    and a few extra divs to make for an enjoyable text editing experience.
 *
 * @customElement
 * @tagname light-editor
 *
 * @event {Event} light-input - Re-emits the textarea's "input" event
 * @event {Event} light-change - Re-emits the textarea's "change" event
 * @event {Event} light-selectionchange - Re-emits the textarea's "selectionchange" event
 * @event {Event} light-focus - Re-emits the textarea's "focus" event
 * @event {Event} light-blur - Re-emits the textarea's "blur" event
 * @event {Event} light-resize - Is emitting whenever the editor resizes.
 *
 */
export default class LightEditor extends BaseElement {
    static styles: import("lit").CSSResult[];
    static properties: {
        value: {
            attribute: boolean;
        };
        initialValue: {
            attribute: string;
        };
        language: {
            reflect: boolean;
        };
        hasInteracted: {
            type: BooleanConstructor;
            attribute: string;
            reflect: boolean;
        };
        preserveWhitespace: {
            type: BooleanConstructor;
            reflect: boolean;
            attribute: string;
        };
    };
    /**
     * The language used for highlighting. Default is "html". "css" and "js" also included by default.
     * @type {string}
     */
    language: string;
    /**
     * The value to set for the text editor
     * @type {string}
     */
    value: string;
    /**
     * This is the value attribute. This is used for resetting the form input.
     * @type {string}
     */
    initialValue: string;
    /**
     * The underlying textarea
     * @type {null | HTMLTextAreaElement}
     */
    textarea: null | HTMLTextAreaElement;
    /**
     * Tracks if the user has interacted with the `<textarea>`
     * @type {boolean}
     */
    hasInteracted: boolean;
    /**
     * Whether to strip whitespace before first character, and after the last character.
     * @type {boolean}
     */
    preserveWhitespace: boolean;
    /**
     * @param {import("lit").PropertyValues<this>} changedProperties
     */
    willUpdate(changedProperties: import("lit").PropertyValues<this>): void;
    click(): void;
    /**
     * @param {FocusOptions} [options]
     */
    focus(options?: FocusOptions | undefined): void;
    render(): import("lit").TemplateResult<1>;
    /**
     * @ignore
     * @param {ResizeObserverEntry[]} entries
     */
    handleTextAreaResize(entries: ResizeObserverEntry[]): void;
    /**
     * @param {import("lit").PropertyValues<this>} changedProperties
     */
    updated(changedProperties: import("lit").PropertyValues<this>): void;
    /**
     * Syncs the `<pre>` element scroll position to the same as the `<textarea>`
     * @internal
     */
    syncScroll(): void;
    /**
     * @ignore
     * @param {Element | undefined} element
     */
    textareaChanged(element: Element | undefined): void;
    textareaResizeObserver: ResizeObserver | undefined;
    /**
     * @ignore
     * @param {Event} e
     */
    handleSlotChangeEvent(e: Event): void;
    /**
     * @ignore
     * @param {KeyboardEvent} evt
     */
    keyupHandler(evt: KeyboardEvent): void;
    /**
     * @ignore
     * @param {KeyboardEvent} evt
     */
    keydownHandler(evt: KeyboardEvent): void;
    setCurrentLineHighlight(): void;
    currentLineNumber: any;
    /**
     * @ignore
     * Override this to use a highlighter of your choice.
     * @param {{code: string, language: string}} options
     */
    highlightCode(options: {
        code: string;
        language: string;
    }): string;
    injectGutterCells(): void;
    getCurrentLineNumber(): number | undefined;
    renderGutterCells(): unknown;
    /**
     * @ignore
     * @param {string} text
     */
    unescapeTags(text: string): string;
    /**
     * @ignore
     * Highlighters strip newlines. But you can see new lines in <textarea>, this fixes that.
     * @param {string} text
     */
    injectNewLine(text: string): string;
}
import { BaseElement } from "../internal/base-element.js";
